# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM postprocessing line plot ViewProvider for the document object"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package view_post_lineplot
#  \ingroup FEM
#  \brief view provider for post line plot object

import FreeCAD
import FreeCADGui

import Plot
from PySide import QtGui, QtCore
from PySide.QtCore import QT_TRANSLATE_NOOP


import io
import numpy as np
import matplotlib as mpl

from vtkmodules.numpy_interface.dataset_adapter import VTKArray

from . import view_base_fempostextractors
from . import view_base_fempostvisualization
from femtaskpanels import task_post_lineplot

from . import view_base_femobject

_GuiPropHelper = view_base_femobject._GuiPropHelper


class EditViewWidget(QtGui.QWidget):

    def __init__(self, obj, post_dialog):
        super().__init__()

        self._object = obj
        self._post_dialog = post_dialog

        # load the ui and set it up
        self.widget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/PostLineplotFieldViewEdit.ui"
        )
        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.widget)
        self.setLayout(layout)

        self.__init_widget()

    def __init_widget(self):
        vobj = self._object.ViewObject

        self.widget.Legend.setText(vobj.Legend)
        self._post_dialog._enumPropertyToCombobox(vobj, "LineStyle", self.widget.LineStyle)
        self._post_dialog._enumPropertyToCombobox(vobj, "MarkerStyle", self.widget.MarkerStyle)
        self.widget.LineWidth.setValue(vobj.LineWidth)
        self.widget.MarkerSize.setValue(vobj.MarkerSize)

        self._setup_color_button(self.widget.Color, vobj.Color, self.colorChanged)

        self.widget.Legend.editingFinished.connect(self.legendChanged)
        self.widget.MarkerStyle.activated.connect(self.markerStyleChanged)
        self.widget.LineStyle.activated.connect(self.lineStyleChanged)
        self.widget.MarkerSize.valueChanged.connect(self.markerSizeChanged)
        self.widget.LineWidth.valueChanged.connect(self.lineWidthChanged)

        # sometimes weird sizes occur with spinboxes
        self.widget.MarkerSize.setMaximumHeight(self.widget.MarkerStyle.sizeHint().height())
        self.widget.LineWidth.setMaximumHeight(self.widget.LineStyle.sizeHint().height())

    def _setup_color_button(self, button, fcColor, callback):

        barColor = QtGui.QColor(*[v * 255 for v in fcColor])
        icon_size = button.iconSize()
        icon_size.setWidth(icon_size.width() * 2)
        button.setIconSize(icon_size)
        pixmap = QtGui.QPixmap(icon_size)
        pixmap.fill(barColor)
        button.setIcon(pixmap)

        action = QtGui.QWidgetAction(button)
        diag = QtGui.QColorDialog(barColor, parent=button)
        diag.setOption(QtGui.QColorDialog.DontUseNativeDialog, True)
        diag.accepted.connect(action.trigger)
        diag.rejected.connect(action.trigger)
        diag.colorSelected.connect(callback)

        action.setDefaultWidget(diag)
        button.addAction(action)
        button.setPopupMode(QtGui.QToolButton.InstantPopup)

    @QtCore.Slot(QtGui.QColor)
    def colorChanged(self, color):

        pixmap = QtGui.QPixmap(self.widget.Color.iconSize())
        pixmap.fill(color)
        self.widget.Color.setIcon(pixmap)

        self._object.ViewObject.Color = color.getRgb()

    @QtCore.Slot(float)
    def lineWidthChanged(self, value):
        self._object.ViewObject.LineWidth = value

    @QtCore.Slot(float)
    def markerSizeChanged(self, value):
        self._object.ViewObject.MarkerSize = value

    @QtCore.Slot(int)
    def markerStyleChanged(self, index):
        self._object.ViewObject.MarkerStyle = index

    @QtCore.Slot(int)
    def lineStyleChanged(self, index):
        self._object.ViewObject.LineStyle = index

    @QtCore.Slot()
    def legendChanged(self):
        self._object.ViewObject.Legend = self.widget.Legend.text()


class EditFieldAppWidget(QtGui.QWidget):

    def __init__(self, obj, post_dialog):
        super().__init__()

        self._object = obj
        self._post_dialog = post_dialog

        # load the ui and set it up
        self.widget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/PostLineplotFieldAppEdit.ui"
        )
        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.widget)
        self.setLayout(layout)

        self.__init_widget()

    def __init_widget(self):
        # set the other properties

        self._post_dialog._enumPropertyToCombobox(self._object, "XField", self.widget.XField)
        self._post_dialog._enumPropertyToCombobox(
            self._object, "XComponent", self.widget.XComponent
        )
        self._post_dialog._enumPropertyToCombobox(self._object, "YField", self.widget.YField)
        self._post_dialog._enumPropertyToCombobox(
            self._object, "YComponent", self.widget.YComponent
        )
        self.widget.Extract.setChecked(self._object.ExtractFrames)

        self.widget.XField.activated.connect(self.xFieldChanged)
        self.widget.XComponent.activated.connect(self.xComponentChanged)
        self.widget.YField.activated.connect(self.yFieldChanged)
        self.widget.YComponent.activated.connect(self.yComponentChanged)
        self.widget.Extract.toggled.connect(self.extractionChanged)

    @QtCore.Slot(int)
    def xFieldChanged(self, index):
        self._object.XField = index
        self._post_dialog._enumPropertyToCombobox(
            self._object, "XComponent", self.widget.XComponent
        )
        self._post_dialog._recompute()

    @QtCore.Slot(int)
    def xComponentChanged(self, index):
        self._object.XComponent = index
        self._post_dialog._recompute()

    @QtCore.Slot(int)
    def yFieldChanged(self, index):
        self._object.YField = index
        self._post_dialog._enumPropertyToCombobox(
            self._object, "YComponent", self.widget.YComponent
        )
        self._post_dialog._recompute()

    @QtCore.Slot(int)
    def yComponentChanged(self, index):
        self._object.YComponent = index
        self._post_dialog._recompute()

    @QtCore.Slot(bool)
    def extractionChanged(self, extract):
        self._object.ExtractFrames = extract
        self._post_dialog._recompute()


class EditIndexAppWidget(QtGui.QWidget):

    def __init__(self, obj, post_dialog):
        super().__init__()

        self._object = obj
        self._post_dialog = post_dialog

        # load the ui and set it up
        self.widget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/PostLineplotIndexAppEdit.ui"
        )
        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.widget)
        self.setLayout(layout)

        self.__init_widget()

    def __init_widget(self):
        # set the other properties

        self.widget.Index.setValue(self._object.Index)
        self._post_dialog._enumPropertyToCombobox(self._object, "YField", self.widget.YField)
        self._post_dialog._enumPropertyToCombobox(
            self._object, "YComponent", self.widget.YComponent
        )

        self.widget.Index.valueChanged.connect(self.indexChanged)
        self.widget.YField.activated.connect(self.yFieldChanged)
        self.widget.YComponent.activated.connect(self.yComponentChanged)

        # sometimes weird sizes occur with spinboxes
        self.widget.Index.setMaximumHeight(self.widget.YField.sizeHint().height())

    @QtCore.Slot(int)
    def indexChanged(self, value):
        self._object.Index = value
        self._post_dialog._recompute()

    @QtCore.Slot(int)
    def yFieldChanged(self, index):
        self._object.YField = index
        self._post_dialog._enumPropertyToCombobox(
            self._object, "YComponent", self.widget.YComponent
        )
        self._post_dialog._recompute()

    @QtCore.Slot(int)
    def yComponentChanged(self, index):
        self._object.YComponent = index
        self._post_dialog._recompute()


class VPPostLineplotFieldData(view_base_fempostextractors.VPPostExtractor):
    """
    A View Provider for extraction of 2D field data specially for histograms
    """

    def __init__(self, vobj):
        super().__init__(vobj)
        vobj.Proxy = self

    def _get_properties(self):

        prop = [
            _GuiPropHelper(
                type="App::PropertyString",
                name="Legend",
                group="Lineplot",
                doc=QT_TRANSLATE_NOOP("FEM", "The name used in the plots legend"),
                value="",
            ),
            _GuiPropHelper(
                type="App::PropertyColor",
                name="Color",
                group="Lineplot",
                doc=QT_TRANSLATE_NOOP("FEM", "The color the line and the markers are drawn with"),
                value=(0, 85, 255, 255),
            ),
            _GuiPropHelper(
                type="App::PropertyEnumeration",
                name="LineStyle",
                group="Lineplot",
                doc=QT_TRANSLATE_NOOP("FEM", "The style the line is drawn in"),
                value=["-", "--", "-.", ":", "None"],
            ),
            _GuiPropHelper(
                type="App::PropertyFloatConstraint",
                name="LineWidth",
                group="Lineplot",
                doc=QT_TRANSLATE_NOOP("FEM", "The width the line is drawn with"),
                value=(1, 0.1, 99, 0.1),
            ),
            _GuiPropHelper(
                type="App::PropertyEnumeration",
                name="MarkerStyle",
                group="Lineplot",
                doc=QT_TRANSLATE_NOOP("FEM", "The style the data markers are drawn with"),
                value=["None", "*", "+", "s", ".", "o", "x"],
            ),
            _GuiPropHelper(
                type="App::PropertyFloatConstraint",
                name="MarkerSize",
                group="Lineplot",
                doc=QT_TRANSLATE_NOOP("FEM", "The size the data markers are drawn in"),
                value=(5, 0.1, 99, 0.1),
            ),
        ]
        return super()._get_properties() + prop

    def attach(self, vobj):
        self.Object = vobj.Object
        self.ViewObject = vobj

    def getIcon(self):
        return ":/icons/FEM_PostField.svg"

    def get_app_edit_widget(self, post_dialog):
        return EditFieldAppWidget(self.Object, post_dialog)

    def get_view_edit_widget(self, post_dialog):
        return EditViewWidget(self.Object, post_dialog)

    def get_preview(self):
        # Returns the preview tuple of icon and label: (QPixmap, str)
        # Note: QPixmap in ratio 2:1

        fig = mpl.pyplot.figure(figsize=(0.2, 0.1), dpi=1000)
        ax = mpl.pyplot.Axes(fig, [0.0, 0.0, 1.0, 1.0])
        ax.set_axis_off()
        fig.add_axes(ax)
        kwargs = self.get_kw_args()
        kwargs["markevery"] = [1]
        ax.plot([0, 0.5, 1], [0.5, 0.5, 0.5], **kwargs)
        data = io.BytesIO()
        mpl.pyplot.savefig(data, bbox_inches=0, transparent=True)
        mpl.pyplot.close()

        pixmap = QtGui.QPixmap()
        pixmap.loadFromData(data.getvalue())

        return (pixmap, self.ViewObject.Legend)

    def get_kw_args(self):
        # builds kw args from the properties
        kwargs = {}

        # colors need a workaround, some error occurs with rgba tuple
        kwargs["color"] = self.ViewObject.Color
        kwargs["markeredgecolor"] = self.ViewObject.Color
        kwargs["markerfacecolor"] = self.ViewObject.Color
        kwargs["linestyle"] = self.ViewObject.LineStyle
        kwargs["linewidth"] = self.ViewObject.LineWidth
        kwargs["marker"] = self.ViewObject.MarkerStyle
        kwargs["markersize"] = self.ViewObject.MarkerSize
        return kwargs

    def get_default_color_property(self):
        return "Color"


class VPPostLineplotIndexOverFrames(VPPostLineplotFieldData):
    """
    A View Provider for extraction of 2D index over frames data
    """

    def __init__(self, vobj):
        super().__init__(vobj)

    def getIcon(self):
        return ":/icons/FEM_PostIndex.svg"

    def get_app_edit_widget(self, post_dialog):
        return EditIndexAppWidget(self.Object, post_dialog)


class VPPostLineplot(view_base_fempostvisualization.VPPostVisualization):
    """
    A View Provider for Lineplot plots
    """

    def __init__(self, vobj):
        super().__init__(vobj)

    def _get_properties(self):

        prop = [
            _GuiPropHelper(
                type="App::PropertyBool",
                name="Grid",
                group="Lineplot",
                doc=QT_TRANSLATE_NOOP(
                    "FEM", "If be the bars should show the cumulative sum left to right"
                ),
                value=True,
            ),
            _GuiPropHelper(
                type="App::PropertyEnumeration",
                name="Scale",
                group="Lineplot",
                doc=QT_TRANSLATE_NOOP("FEM", "The scale the axis are drawn in"),
                value=["linear", "semi-log x", "semi-log y", "log"],
            ),
            _GuiPropHelper(
                type="App::PropertyString",
                name="Title",
                group="Plot",
                doc=QT_TRANSLATE_NOOP("FEM", "The histogram plot title"),
                value="",
            ),
            _GuiPropHelper(
                type="App::PropertyString",
                name="XLabel",
                group="Plot",
                doc=QT_TRANSLATE_NOOP("FEM", "The label shown for the histogram X axis"),
                value="",
            ),
            _GuiPropHelper(
                type="App::PropertyString",
                name="YLabel",
                group="Plot",
                doc=QT_TRANSLATE_NOOP("FEM", "The label shown for the histogram Y axis"),
                value="",
            ),
            _GuiPropHelper(
                type="App::PropertyBool",
                name="Legend",
                group="Plot",
                doc=QT_TRANSLATE_NOOP("FEM", "Determines if the legend is plotted"),
                value=True,
            ),
            _GuiPropHelper(
                type="App::PropertyEnumeration",
                name="LegendLocation",
                group="Plot",
                doc=QT_TRANSLATE_NOOP("FEM", "Determines if the legend is plotted"),
                value=[
                    "best",
                    "upper right",
                    "upper left",
                    "lower left",
                    "lower right",
                    "right",
                    "center left",
                    "center right",
                    "lower center",
                    "upper center",
                    "center",
                ],
            ),
        ]
        return prop

    def getIcon(self):
        return ":/icons/FEM_PostLineplot.svg"

    def setEdit(self, vobj, mode):

        # build up the task panel
        taskd = task_post_lineplot._TaskPanel(vobj)

        # show it
        FreeCADGui.Control.showDialog(taskd)

        return True

    def show_visualization(self):

        if not hasattr(self, "_plot") or not self._plot:
            main = FreeCADGui.getMainWindow()
            self._plot = Plot.Plot()
            self._plot.setWindowTitle(self.Object.Label)
            self._plot.setParent(main)
            self._plot.setWindowFlags(QtGui.Qt.Dialog)
            self._plot.resize(
                main.size().height() / 2, main.size().height() / 3
            )  # keep the aspect ratio
            self.update_visualization()

        self._plot.show()

    def get_kw_args(self, obj):
        view = obj.ViewObject
        if not view or not hasattr(view, "Proxy"):
            return {}
        if not hasattr(view.Proxy, "get_kw_args"):
            return {}
        return view.Proxy.get_kw_args()

    def update_visualization(self):

        if not hasattr(self, "_plot") or not self._plot:
            return

        self._plot.axes.clear()

        # we do not iterate the table, but iterate the children. This makes it possible
        # to attribute the correct styles
        plotted = False
        for child in self.Object.Group:

            table = child.Table
            kwargs = self.get_kw_args(child)

            # iterate over the table and plot all (note: column 0 is always X!)
            color_factor = np.linspace(1, 0.5, int(table.GetNumberOfColumns() / 2))
            legend_multiframe = table.GetNumberOfColumns() > 2

            for i in range(0, table.GetNumberOfColumns(), 2):

                plotted = True

                # add the kw args, with some slide change over color for multiple frames
                tmp_args = {}
                for key in kwargs:
                    if "color" in key:
                        value = np.array(kwargs[key]) * color_factor[int(i / 2)]
                        tmp_args[key] = mpl.colors.to_hex(value)
                    else:
                        tmp_args[key] = kwargs[key]

                xdata = VTKArray(table.GetColumn(i))
                ydata = VTKArray(table.GetColumn(i + 1))

                # ensure points are visible if it is a single datapoint
                if len(xdata) == 1 and tmp_args["marker"] == "None":
                    tmp_args["marker"] = "o"

                # legend labels
                if child.ViewObject.Legend:
                    if not legend_multiframe:
                        label = child.ViewObject.Legend
                    else:
                        postfix = table.GetColumnName(i + 1).split("-")[-1]
                        label = child.ViewObject.Legend + " - " + postfix
                else:
                    legend_prefix = ""
                    if len(self.Object.Group) > 1:
                        legend_prefix = child.Source.Label + ": "
                    label = legend_prefix + table.GetColumnName(i + 1)

                match self.ViewObject.Scale:
                    case "log":
                        self._plot.axes.loglog(xdata, ydata, **tmp_args, label=label)
                    case "semi-log x":
                        self._plot.axes.semilogx(xdata, ydata, **tmp_args, label=label)
                    case "semi-log y":
                        self._plot.axes.semilogy(xdata, ydata, **tmp_args, label=label)
                    case _:
                        self._plot.axes.plot(xdata, ydata, **tmp_args, label=label)

        if self.ViewObject.Title:
            self._plot.axes.set_title(self.ViewObject.Title)
        if self.ViewObject.XLabel:
            self._plot.axes.set_xlabel(self.ViewObject.XLabel)
        if self.ViewObject.YLabel:
            self._plot.axes.set_ylabel(self.ViewObject.YLabel)

        if self.ViewObject.Legend and plotted:
            self._plot.axes.legend(loc=self.ViewObject.LegendLocation)

        self._plot.axes.grid(self.ViewObject.Grid)
        self._plot.update()

    def get_next_default_color(self):
        # we use the next color in order. We do not check (yet) if this
        # color is already taken
        i = len(self.Object.Group)
        cmap = mpl.pyplot.get_cmap("tab10")
        return cmap(i)



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM postprocessing line plot ViewProvider for the document object"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package view_post_lineplot
#  \ingroup FEM
#  \brief view provider for post line plot object

import FreeCAD
import FreeCADGui

from PySide import QtGui

from femtaskpanels import task_post_extractor


class VPPostExtractor:
    """
    A View Provider for extraction of data
    """

    def __init__(self, vobj):
        vobj.Proxy = self
        self._setup_properties(vobj)

    def _setup_properties(self, vobj):
        pl = vobj.PropertiesList
        for prop in self._get_properties():
            if not prop.name in pl:
                prop.add_to_object(vobj)

    def _get_properties(self):
        return []

    def attach(self, vobj):
        self.Object = vobj.Object  # used on various places, claim childreens, get icon, etc.
        self.ViewObject = vobj

    def isShow(self):
        return True

    def onChanged(self, vobj, prop):

        # one of our view properties was changed. Lets inform our parent visualization
        # that this happened, as this is the one that needs to redraw

        if prop == "Proxy":
            return

        group = vobj.Object.getParentGroup()
        if not group:
            return

        if hasattr(group.ViewObject, "Proxy") and hasattr(
            group.ViewObject.Proxy, "childViewPropertyChanged"
        ):

            group.ViewObject.Proxy.childViewPropertyChanged(vobj, prop)

    def setEdit(self, vobj, mode):

        # build up the task panel
        taskd = task_post_extractor._ExtractorTaskPanel(vobj.Object)

        # show it
        FreeCADGui.Control.showDialog(taskd)

        return True

    def unsetEdit(self, vobj, mode=0):
        FreeCADGui.Control.closeDialog()
        return True

    def doubleClicked(self, vobj):
        guidoc = FreeCADGui.getDocument(vobj.Object.Document)

        # check if another VP is in edit mode and close it then
        if guidoc.getInEdit():
            FreeCADGui.Control.closeDialog()
            guidoc.resetEdit()

        guidoc.setEdit(vobj.Object.Name)

        return True

    def dumps(self):
        return None

    def loads(self, state):
        return None

    # To be implemented by subclasses:
    # ################################

    def get_default_color_property(self):
        # Returns the property name to set the default color to.
        # Return None if no such property
        raise FreeCAD.Base.FreeCADError("Not implemented")

    def get_default_field_properties(self):
        # Returns the property name to which the default field name should be set
        # ret: [FieldProperty, ComponentProperty]
        raise FreeCAD.Base.FreeCADError("Not implemented")

    def get_kw_args(self):
        # Returns the matplotlib plot keyword arguments that represent the
        # properties of the object.
        raise FreeCAD.Base.FreeCADError("Not implemented")

    def get_app_edit_widget(self, post_dialog):
        # Returns a widgets for editing the object (not viewprovider!)
        # The widget will be part of the provided post_dialog, and
        # should use its functionality to inform of changes.
        raise FreeCAD.Base.FreeCADError("Not implemented")

    def get_view_edit_widget(self, post_dialog):
        # Returns a widgets for editing the viewprovider (not object!)
        # The widget will be part of the provided post_dialog, and
        # should use its functionality to inform of changes.
        raise FreeCAD.Base.FreeCADError("Not implemented")

    def get_preview(self):
        # Returns the preview tuple of icon and label: (QPixmap, str)
        # Note: QPixmap in ratio 2:1
        raise FreeCAD.Base.FreeCADError("Not implemented")



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM postprocessing glyph filter ViewProvider for the document object"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package view_post_glyphfilter
#  \ingroup FEM
#  \brief view provider for post glyph filter object

import FreeCAD
import FreeCADGui

import FemGui
from PySide import QtGui
from femtaskpanels import task_post_glyphfilter


class VPPostGlyphFilter:
    """
    A View Provider for the PostGlyphFilter object
    """

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        return ":/icons/FEM_PostFilterGlyph.svg"

    def getDisplayModes(self, obj):
        # Mandatory, as the ViewProviderPostFilterPython does not add any
        # display modes. We can choose here any that is supported by it:
        # "Outline", "Nodes", "Surface", "Surface with Edges",
        # "Wireframe", "Wireframe (surface only)", "Nodes (surface only)"

        # only surface makes sense for the glyphs
        return ["Surface"]

    def setDisplayMode(self, mode):
        # the post object viewprovider implements the different display modes
        # via vtk filter, not via masking modes. Hence we need to make sure
        # to always stay in the "Default" masking mode, no matter the display mode
        return "Default"

    def setEdit(self, vobj, mode):
        # make sure we see what we edit
        vobj.show()

        # build up the task panel
        taskd = task_post_glyphfilter._TaskPanel(vobj)

        # show it
        FreeCADGui.Control.showDialog(taskd)

        return True

    def unsetEdit(self, vobj, mode):
        FreeCADGui.Control.closeDialog()
        return True

    def dumps(self):
        return None

    def loads(self, state):
        return None



# SPDX-License-Identifier: LGPL-2.1-or-later

import FreeCAD
import math, os, unittest


class TestSketchCarbonCopyReverseMapping(unittest.TestCase):
    def setUp(self):
        location = os.path.dirname(os.path.realpath(__file__))
        self.Doc = FreeCAD.openDocument(
            os.path.join(location, "TestSketchCarbonCopyReverseMapping.FCStd"), True
        )

    def test_CarbonCopyReverseMapping(self):
        r = 10.0
        rad30 = math.radians(30)
        cos30 = math.cos(rad30)
        sin30 = math.sin(rad30)

        expected_geometries_pattern_001 = {
            "Sketch001": {
                0: {"start": FreeCAD.Vector(10, 10, 0), "end": FreeCAD.Vector(20, 10, 0)},
                1: {
                    "start": FreeCAD.Vector(10, 10, 0),
                    "end": FreeCAD.Vector(10 + cos30 * r, 10 + sin30 * r, 0),
                },
                2: {
                    "start": FreeCAD.Vector(10, 10, 0),
                    "end": FreeCAD.Vector(10 + sin30 * r, 10 + cos30 * r, 0),
                },
                3: {
                    "start": FreeCAD.Vector(10, 10, 0),
                    "end": FreeCAD.Vector(10 - sin30 * r, 10 + cos30 * r, 0),
                },
                4: {"location": FreeCAD.Vector(-10, 10, 0)},
                5: {"location": FreeCAD.Vector(10, -10, 0)},
                6: {
                    "start": FreeCAD.Vector(5, 15, 0),
                    "end": FreeCAD.Vector(5 + sin30 * r, 15 + cos30 * r, 0),
                },
                7: {
                    "start": FreeCAD.Vector(15, 15, 0),
                    "end": FreeCAD.Vector(15 - sin30 * r, 15 + cos30 * r, 0),
                },
            },
            "Sketch002": {
                0: {"start": FreeCAD.Vector(-10, 10, 0), "end": FreeCAD.Vector(-20, 10, 0)},
                1: {
                    "start": FreeCAD.Vector(-10, 10, 0),
                    "end": FreeCAD.Vector(-10 - cos30 * r, 10 + sin30 * r, 0),
                },
                2: {
                    "start": FreeCAD.Vector(-10, 10, 0),
                    "end": FreeCAD.Vector(-10 - sin30 * r, 10 + cos30 * r, 0),
                },
                3: {
                    "start": FreeCAD.Vector(-10, 10, 0),
                    "end": FreeCAD.Vector(-10 + sin30 * r, 10 + cos30 * r, 0),
                },
                4: {"location": FreeCAD.Vector(10, 10, 0)},
                5: {"location": FreeCAD.Vector(-10, -10, 0)},
                6: {
                    "start": FreeCAD.Vector(-5, 15, 0),
                    "end": FreeCAD.Vector(-5 - sin30 * r, 15 + cos30 * r, 0),
                },
                7: {
                    "start": FreeCAD.Vector(-15, 15, 0),
                    "end": FreeCAD.Vector(-15 + sin30 * r, 15 + cos30 * r, 0),
                },
            },
            "Sketch003": {
                0: {"start": FreeCAD.Vector(10, -10, 0), "end": FreeCAD.Vector(20, -10, 0)},
                1: {
                    "start": FreeCAD.Vector(10, -10, 0),
                    "end": FreeCAD.Vector(10 + cos30 * r, -10 - sin30 * r, 0),
                },
                2: {
                    "start": FreeCAD.Vector(10, -10, 0),
                    "end": FreeCAD.Vector(10 + sin30 * r, -10 - cos30 * r, 0),
                },
                3: {
                    "start": FreeCAD.Vector(10, -10, 0),
                    "end": FreeCAD.Vector(10 - sin30 * r, -10 - cos30 * r, 0),
                },
                4: {"location": FreeCAD.Vector(-10, -10, 0)},
                5: {"location": FreeCAD.Vector(10, 10, 0)},
                6: {
                    "start": FreeCAD.Vector(5, -15, 0),
                    "end": FreeCAD.Vector(5 + sin30 * r, -15 - cos30 * r, 0),
                },
                7: {
                    "start": FreeCAD.Vector(15, -15, 0),
                    "end": FreeCAD.Vector(15 - sin30 * r, -15 - cos30 * r, 0),
                },
            },
            "Sketch004": {
                0: {"start": FreeCAD.Vector(-10, -10, 0), "end": FreeCAD.Vector(-20, -10, 0)},
                1: {
                    "start": FreeCAD.Vector(-10, -10, 0),
                    "end": FreeCAD.Vector(-10 - cos30 * r, -10 - sin30 * r, 0),
                },
                2: {
                    "start": FreeCAD.Vector(-10, -10, 0),
                    "end": FreeCAD.Vector(-10 - sin30 * r, -10 - cos30 * r, 0),
                },
                3: {
                    "start": FreeCAD.Vector(-10, -10, 0),
                    "end": FreeCAD.Vector(-10 + sin30 * r, -10 - cos30 * r, 0),
                },
                4: {"location": FreeCAD.Vector(10, -10, 0)},
                5: {"location": FreeCAD.Vector(-10, 10, 0)},
                6: {
                    "start": FreeCAD.Vector(-5, -15, 0),
                    "end": FreeCAD.Vector(-5 - sin30 * r, -15 - cos30 * r, 0),
                },
                7: {
                    "start": FreeCAD.Vector(-15, -15, 0),
                    "end": FreeCAD.Vector(-15 + sin30 * r, -15 - cos30 * r, 0),
                },
            },
            "Sketch005": {
                0: {"start": FreeCAD.Vector(10, 10, 0), "end": FreeCAD.Vector(20, 10, 0)},
                1: {
                    "start": FreeCAD.Vector(10, 10, 0),
                    "end": FreeCAD.Vector(10 + cos30 * r, 10 + sin30 * r, 0),
                },
                2: {
                    "start": FreeCAD.Vector(10, 10, 0),
                    "end": FreeCAD.Vector(10 + sin30 * r, 10 + cos30 * r, 0),
                },
                3: {
                    "start": FreeCAD.Vector(10, 10, 0),
                    "end": FreeCAD.Vector(10 - sin30 * r, 10 + cos30 * r, 0),
                },
                4: {"location": FreeCAD.Vector(-10, 10, 0)},
                5: {"location": FreeCAD.Vector(10, -10, 0)},
                6: {
                    "start": FreeCAD.Vector(5, 15, 0),
                    "end": FreeCAD.Vector(5 + sin30 * r, 15 + cos30 * r, 0),
                },
                7: {
                    "start": FreeCAD.Vector(15, 15, 0),
                    "end": FreeCAD.Vector(15 - sin30 * r, 15 + cos30 * r, 0),
                },
            },
        }

        expected_geometries_pattern_006 = {
            "Sketch007": {
                0: {"start": FreeCAD.Vector(-35, -15, 0), "end": FreeCAD.Vector(-35, -5, 0)},
                1: {"start": FreeCAD.Vector(-35, -5, 0), "end": FreeCAD.Vector(-20, -5, 0)},
                2: {"start": FreeCAD.Vector(-20, -5, 0), "end": FreeCAD.Vector(-20, -15, 0)},
                3: {"start": FreeCAD.Vector(-20, -15, 0), "end": FreeCAD.Vector(-35, -15, 0)},
            }
        }

        def execute_carbon_copy(src, dist):
            obj = FreeCAD.ActiveDocument.getObject(dist)
            obj.setAllowUnaligned(False)
            obj.carbonCopy(src, False)
            FreeCAD.ActiveDocument.recompute()

        def check_placement(expected_dict):
            for sketch, correct_points in expected_dict.items():
                obj = FreeCAD.ActiveDocument.getObject(sketch)
                for i, pts in correct_points.items():
                    geom = obj.Geometry[i]
                    if "start" in pts:
                        start = geom.StartPoint
                        end = geom.EndPoint
                        self.assertAlmostEqual((start - pts["start"]).Length, 0)
                        self.assertAlmostEqual((end - pts["end"]).Length, 0)
                    elif "location" in pts:
                        loc = geom.Location
                        self.assertAlmostEqual((loc - pts["location"]).Length, 0)

        execute_carbon_copy("Sketch001", "Sketch002")
        execute_carbon_copy("Sketch001", "Sketch003")
        execute_carbon_copy("Sketch001", "Sketch004")
        execute_carbon_copy("Sketch001", "Sketch005")
        execute_carbon_copy("Sketch006", "Sketch007")

        check_placement(expected_geometries_pattern_001)
        check_placement(expected_geometries_pattern_006)

    def tearDown(self):
        FreeCAD.closeDocument(self.Doc.Name)



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM postprocessing ldata view and extraction widget"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package data_extraction
#  \ingroup FEM
#  \brief A widget for data extraction. Used in the PostObject task panel.

from . import vtk_table_view

from PySide import QtCore, QtGui

from vtkmodules.vtkCommonCore import vtkVersion
from vtkmodules.vtkCommonDataModel import vtkTable
from vtkmodules.vtkFiltersGeneral import vtkSplitColumnComponents

from Fem import vtkVersionCheck

if vtkVersionCheck(
    vtkVersion.GetVTKMajorVersion(), vtkVersion.GetVTKMinorVersion()
) >= vtkVersionCheck(9, 3):
    from vtkmodules.vtkFiltersCore import vtkAttributeDataToTableFilter
else:
    from vtkmodules.vtkInfovisCore import vtkDataObjectToTable


import FreeCAD
import FreeCADGui

import femobjects.base_fempostextractors as extr
from femtaskpanels.base_fempostpanel import _BasePostTaskPanel

from . import extract_link_view

ExtractLinkView = extract_link_view.ExtractLinkView


class DataExtraction(_BasePostTaskPanel):
    # The class is not a widget itself, but provides a widget. It implements
    # all required callbacks for the widget and the task dialog.
    # Note: This object is created and used from c++! See PostTaskExtraction

    def __init__(self, vobj):

        super().__init__(vobj.Object)

        self.ViewObject = vobj
        self.Object = vobj.Object

        self.widget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/TaskPostExtraction.ui"
        )

        # connect all signals as required
        self.widget.Data.clicked.connect(self.showData)
        self.widget.Summary.clicked.connect(self.showSummary)

        # setup the data models
        self.data_model = vtk_table_view.VtkTableModel()
        self.summary_model = vtk_table_view.VtkTableSummaryModel()

        # generate the data
        self.onPostDataChanged(self.Object)

        # setup the extraction widget
        self._extraction_view = ExtractLinkView(self.Object, True, self)
        self.widget.layout().addSpacing(self.widget.Data.size().height() / 3)
        self.widget.layout().addWidget(self._extraction_view)
        self._extraction_view.repopulate()

    @QtCore.Slot()
    def showData(self):

        dialog = QtGui.QDialog(self.widget)
        dialog.setWindowTitle(f"Data of {self.Object.Label}")
        widget = vtk_table_view.VtkTableView(self.data_model)
        layout = QtGui.QVBoxLayout()
        layout.addWidget(widget)
        layout.setContentsMargins(0, 0, 0, 0)

        dialog.setLayout(layout)
        dialog.resize(1500, 900)
        dialog.show()

    @QtCore.Slot()
    def showSummary(self):

        dialog = QtGui.QDialog(self.widget)
        dialog.setWindowTitle(f"Data Summary of {self.Object.Label}")
        widget = vtk_table_view.VtkTableView(self.summary_model)
        layout = QtGui.QVBoxLayout()
        layout.addWidget(widget)
        layout.setContentsMargins(0, 0, 0, 0)

        dialog.setLayout(layout)
        dialog.resize(600, 900)
        dialog.show()

    def isGuiTaskOnly(self):
        # If all panels return true it omits the Apply button in the dialog
        return True

    def onPostDataChanged(self, obj):

        algo = obj.getOutputAlgorithm()
        if not algo:
            self.data_model.setTable(vtkTable())

        if vtkVersionCheck(
            vtkVersion.GetVTKMajorVersion(), vtkVersion.GetVTKMinorVersion()
        ) >= vtkVersionCheck(9, 3):
            filter = vtkAttributeDataToTableFilter()
        else:
            filter = vtkDataObjectToTable()
            filter.SetFieldType(vtkDataObjectToTable.POINT_DATA)

        filter.SetInputConnection(0, algo.GetOutputPort(0))
        filter.Update()
        table = filter.GetOutputDataObject(0)

        # add the points
        points = algo.GetOutputDataObject(0).GetPoints().GetData()
        table.AddColumn(points)

        # split the components
        splitter = vtkSplitColumnComponents()
        splitter.SetNamingModeToNamesWithParens()
        splitter.SetInputData(0, table)

        splitter.Update()
        table = splitter.GetOutputDataObject(0)

        self.data_model.setTable(table)
        self.summary_model.setTable(table)

    def apply(self):
        pass

    def initiallyCollapsed(self):
        # if we do not have any extractions to show we hide initially to remove clutter

        for obj in self.Object.InList:
            if extr.is_extractor_object(obj):
                return False

        return True



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Methods to verify if the python VTK module is the correct one

FreeCAD is linked with VTK libraries during its build process. To use the VTK
python module and pass objects between python and c++ the compiled module library
needs to be linked to the exact same vtk library as FreeCAD is. This is ensured by
installing VTK via linux app managers: All known distros install the python side
packages together with vtk libs. Libpack and other OS systems ensure this too.

However, if a vtk python package is installed manually, e.g. by "pip install vtk",
it could be found instead of the system module. This python API brings its own
set of VTK libs, and hence object passing in FreeCAD fails. (Note: import and
pure vtk python code still works, only passing to c++ fails)

This file provides functions that detect this situation and inform the user.
Additionally we try to find the correct module in the path and offer to use
it instead.

Note that this problem occurs with all "compiled binary" python APIs, also
with PySide. It is the users responsibility to handle his python path and keep
it clean/working. The functions provided here are a workaround only.
"""

__title__ = "FEM GUI vtk python module check"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"


# Note: This file is imported from FreeCAD App files. Do not import any FreeCADGui
#       directly to support cmd line use.

__user_input_received = False


def vtk_module_compatible():
    # checks if the VTK library FreeCAD is build against is the one used by
    # the python module

    # make sure we do not contaminate the modules with vtk to not trick
    # the check later
    unload = not _vtk_is_loaded()

    import Fem
    from vtkmodules.vtkCommonCore import vtkVersion, vtkBitArray

    # simple version check
    if Fem.getVtkVersion() != vtkVersion.GetVTKVersion():
        return False

    # check binary compatibility
    result = Fem.isVtkCompatible(vtkBitArray())

    if unload:
        # cleanup our own import
        _unload_vtk_modules()

    return result


def _vtk_is_loaded():
    import sys

    return any("vtkmodules" in module for module in sys.modules)


def _unload_vtk_modules():
    # unloads all loaded vtk modules
    # NOTE: does not remove any stored references in objects

    import sys

    for module in sys.modules.copy():
        if "vtkmodules" in module:
            del sys.modules[module]


def _find_compatible_module():
    # Check all python path folders if they contain a vtk module

    import Fem
    import sys

    # remove module from runtime
    _unload_vtk_modules()

    path = sys.path.copy()

    for folder in reversed(path):
        try:
            # use a single folder as path and try to load vtk
            sys.path = [folder]
            if vtk_module_compatible():
                # we do still unload, to let the user decide if they want to use it
                _unload_vtk_modules()
                sys.path = path
                return folder

        except:
            continue

    # reset the correct path and indicate that we failed
    sys.path = path
    return None


def _load_vtk_from(folder):

    import sys

    path = sys.path
    try:
        sys.path = [folder]
        import vtkmodules
    finally:
        sys.path = path


# If FreeCAD is build with VTK python support this function checks if the
# used python module is compatible with the c++ lib. Does inform the user
# if not so and offers the correct module, if available
#
# Note: Call this also from Python feature module, as on document load
#       this can be loaded before initializing FEM workbench.
def vtk_module_handling():

    import sys
    import FreeCAD

    if not "BUILD_FEM_VTK_PYTHON" in FreeCAD.__cmake__:
        # no VTK python api support in FreeCAD
        return

    # only ask user once per session
    global __user_input_received
    if __user_input_received:
        return
    __user_input_received = True

    loaded = _vtk_is_loaded()

    # check if we are compatible
    if not vtk_module_compatible():

        if not FreeCAD.GuiUp:
            FreeCAD.Console.PrintError(
                "FEM: VTK Python module is not compatible with internal VTK library"
            )
            return

        import FreeCAD, Fem
        from vtkmodules.vtkCommonCore import vtkVersion
        import inspect
        from PySide import QtGui

        translate = FreeCAD.Qt.translate

        path = inspect.getfile(vtkVersion)
        path = path[: path.find("vtkmodules")]

        message = translate(
            "FEM",
            (
                "FreeCAD is linked to a different VTK library than the imported "
                "VTK Python module. This is incompatible and will lead to errors."
                "\n\nWrong python module is imported from: \n{}"
            ),
        ).format(path)

        buttons = QtGui.QMessageBox.Discard

        # check if there is any compatible vtk module
        compatible_module = _find_compatible_module()

        if compatible_module:
            # there is a compatible module of VTK available.
            message += translate("FEM", "\n\nCorrect module found in: \n{}").format(
                compatible_module
            )

            if not loaded:
                # vtk was not loaded beforehand, therefore we can realistically reload
                message += translate("FEM", "\n\nShould this module be loaded instead?")

                buttons = QtGui.QMessageBox.Yes | QtGui.QMessageBox.No

            else:
                message += translate(
                    "FEM",
                    (
                        "\n\nAs the wrong module was already loaded, a reload is not possible. "
                        "Restart FreeCAD to get the option for loading this module."
                    ),
                )

        else:
            message += translate(
                "FEM", "\n\nNo matching module was found in the current Python path."
            )

        # raise a dialog to the user
        import FreeCADGui

        button = QtGui.QMessageBox.critical(
            FreeCADGui.getMainWindow(),
            translate("FEM", "VTK Python module conflict"),
            message,
            buttons=buttons,
        )

        if button == QtGui.QMessageBox.Yes:
            # try to reload the correct vtk module
            _load_vtk_from(compatible_module)


# Returns if vtk python is incompatible and hence operations need to be aborted.
# If inform=True the user gets informed by dialog about incompatibilities
def vtk_compatibility_abort(inform=True):

    if not vtk_module_compatible():

        if inform:
            # raise a dialog to the user that this functionality is not available
            import FreeCAD
            import FreeCADGui
            from PySide import QtGui

            translate = FreeCAD.Qt.translate

            button = QtGui.QMessageBox.critical(
                FreeCADGui.getMainWindow(),
                translate("FEM", "VTK Python Module Conflict"),
                translate(
                    "FEM", "This functionality is not available due to VTK Python module conflict"
                ),
                buttons=QtGui.QMessageBox.Discard,
            )

        return True

    return False



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import FreeCAD
from PySide.QtCore import QT_TRANSLATE_NOOP


class CuttingToolMixin:
    """
    This is a interface class to indicate that the ToolBit can chip, i.e.
    it has a Chipload property.
    It is used to determine if the tool bit can be used for chip removal.
    """

    def __init__(self, obj, *args, **kwargs):
        obj.addProperty(
            "App::PropertyLength",
            "Chipload",
            "Attributes",
            QT_TRANSLATE_NOOP("App::Property", "Chipload per tooth"),
        )
        obj.Chipload = FreeCAD.Units.Quantity("0.0 mm")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides various functions to work with cubic shapes (parallelepipeds)."""
## @package cuboids
# \ingroup draftgeoutils
# \brief Provides various functions for cubic shapes (parallelepipeds).

import math
import lazy_loader.lazy_loader as lz

import FreeCAD as App

from draftgeoutils.general import geomType, vec, precision

# Delay import of module until first use because it is heavy
Part = lz.LazyLoader("Part", globals(), "Part")

## \addtogroup draftgeoutils
# @{


def isCubic(shape):
    """Return True if the shape is a parallelepiped (cuboid).

    A parallelepiped of cube-like shape has 8 vertices, 6 faces, 12 edges,
    and all angles are 90 degrees between its edges.
    """
    # first we try fast methods
    if len(shape.Vertexes) != 8 or len(shape.Faces) != 6 or len(shape.Edges) != 12:
        return False

    for e in shape.Edges:
        if geomType(e) != "Line":
            return False

    # if ok until now, let's do more advanced testing
    for f in shape.Faces:
        if len(f.Edges) != 4:
            return False

        for i in range(4):
            e1 = vec(f.Edges[i])
            if i < 3:
                e2 = vec(f.Edges[i + 1])
            else:
                e2 = vec(f.Edges[0])
            rpi = [0.0, round(math.pi / 2, precision())]
            if round(e1.getAngle(e2), precision()) not in rpi:
                return False

    return True


def getCubicDimensions(shape):
    """Return a list containing the placement, and dimensions of the shape.

    The dimensions are length, width and height of a the parallelepiped, rounded to
    the value indicated by `precision`. With the length larger than or equal to width.
    The placement point is the center of the lowest face.

    If it is not a parallelepiped (cuboid), return None.
    """

    def _edge_vector(base_point, edge):
        """Get a unit-vector from the vertices of an edge. The base_point
        must match one of these points and determines the start point."""
        points = [v.Point for v in edge.Vertexes]
        if points[1].isEqual(base_point, 1e-6):
            points.reverse()
        return (points[1] - points[0]).normalize()

    if not isCubic(shape):
        return None

    # determine lowest face, which will be our base
    z = [10, 1000000000000]
    for i in range(len(shape.Faces)):
        if shape.Faces[i].CenterOfMass.z < z[1]:
            z = [i, shape.Faces[i].CenterOfMass.z]

    if z[0] > 5:
        return None

    base_face = shape.Faces[z[0]]
    base_norm = base_face.normalAt(0, 0)
    base_vertex = base_face.Vertexes[0]
    corner_edges = shape.ancestorsOfType(base_vertex, Part.Edge)
    edge_z = None
    for edge in corner_edges:
        edge_vector = vec(edge).normalize()
        if edge_vector.isEqual(base_norm, 1e-6) or edge_vector.isEqual(-base_norm, 1e-6):
            edge_z = edge
            break

    if edge_z is None:
        return None

    corner_edges.remove(edge_z)
    edge_x, edge_y = corner_edges
    if edge_x.Length < edge_y.Length:
        edge_x, edge_y = edge_y, edge_x

    base_point = base_vertex.Point
    rot = App.Rotation(
        _edge_vector(base_point, edge_x),
        _edge_vector(base_point, edge_y),
        _edge_vector(base_point, edge_z),
        "ZXY",
    )

    return [
        App.Placement(base_face.CenterOfMass, rot),
        round(edge_x.Length, precision()),
        round(edge_y.Length, precision()),
        round(edge_z.Length, precision()),
    ]


## @}



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD visualization registry"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package post_visualization
#  \ingroup FEM
#  \brief A registry to collect visualizations for use in menus

# Note: This file is imported from FreeCAD App files. Do not import any FreeCADGui
#       directly to support cmd line use.

import copy
from dataclasses import dataclass

from PySide import QtCore

import FreeCAD


# Registry to handle visualization commands
# #########################################

_registry = {}


@dataclass
class _Extraction:

    name: str
    icon: str
    dimension: str
    extracttype: str
    module: str
    factory: str


@dataclass
class _Visualization:

    name: str
    icon: str
    module: str
    factory: str
    extractions: list[_Extraction]


# Register a visualization by type, icon and factory function
def register_visualization(visualization_type, icon, module, factory):
    if visualization_type in _registry:
        raise ValueError("Visualization type already registered")

    _registry[visualization_type] = _Visualization(visualization_type, icon, module, factory, [])


def register_extractor(
    visualization_type, extraction_type, icon, dimension, etype, module, factory
):

    if not visualization_type in _registry:
        raise ValueError("visualization not registered yet")

    extraction = _Extraction(extraction_type, icon, dimension, etype, module, factory)
    _registry[visualization_type].extractions.append(extraction)


def get_registered_visualizations():
    return copy.deepcopy(_registry)


def _to_command_name(name):
    return "FEM_PostVisualization" + name


class _VisualizationGroupCommand:

    def GetCommands(self):
        visus = _registry.keys()
        cmds = [_to_command_name(v) for v in visus]
        return cmds

    def GetDefaultCommand(self):
        return 0

    def GetResources(self):
        return {
            "MenuText": QtCore.QT_TRANSLATE_NOOP("FEM", "Data Visualizations"),
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(
                "FEM", "Different visualizations to show post processing data in"
            ),
        }

    def IsActive(self):
        if not FreeCAD.ActiveDocument:
            return False

        import FemGui

        return bool(FemGui.getActiveAnalysis())


class _VisualizationCommand:

    def __init__(self, visualization_type):
        self._visualization_type = visualization_type

    def GetResources(self):

        cmd = _to_command_name(self._visualization_type)
        vis = _registry[self._visualization_type]
        tooltip = f"Create a {self._visualization_type} post processing data visualization"

        return {
            "Pixmap": vis.icon,
            "MenuText": QtCore.QT_TRANSLATE_NOOP(cmd, "Create {}".format(self._visualization_type)),
            "Accel": "",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(cmd, tooltip),
            "CmdType": "AlterDoc",
        }

    def IsActive(self):
        # active analysis available
        if not FreeCAD.ActiveDocument:
            return False

        import FemGui

        return bool(FemGui.getActiveAnalysis())

    def Activated(self):
        import FreeCADGui

        vis = _registry[self._visualization_type]
        FreeCAD.ActiveDocument.openTransaction(f"Create {vis.name}")

        FreeCADGui.addModule(vis.module)
        FreeCADGui.addModule("FemGui")

        FreeCADGui.doCommand(f"obj = {vis.module}.{vis.factory}(FreeCAD.ActiveDocument)")
        FreeCADGui.doCommand(f"FemGui.getActiveAnalysis().addObject(obj)")

        FreeCADGui.Selection.clearSelection()
        FreeCADGui.doCommand("FreeCADGui.ActiveDocument.setEdit(obj)")


def setup_commands(toplevel_name):
    # creates all visualization commands and registers them. The
    # toplevel group command will have the name provided to this function.

    import FreeCADGui

    # first all visualization and extraction commands
    for vis in _registry:
        FreeCADGui.addCommand(_to_command_name(vis), _VisualizationCommand(vis))

    # build the group command!
    FreeCADGui.addCommand("FEM_PostVisualization", _VisualizationGroupCommand())



# ***************************************************************************
# *   Copyright (c) 2016 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM mesh boundary layer task panel for the document object"
__author__ = "Bernd Hahnebach, Qingfeng Xia"
__url__ = "https://www.freecad.org"

## @package task_mesh_boundarylayer
#  \ingroup FEM
#  \brief task panel for mesh boundary object

from PySide import QtCore

import FreeCAD
import FreeCADGui

from femguiutils import selection_widgets
from . import base_femtaskpanel


class _TaskPanel(base_femtaskpanel._BaseTaskPanel):
    """
    The TaskPanel for editing References property of MeshBoundaryLayer objects
    """

    def __init__(self, obj):
        super().__init__(obj)

        # parameter widget
        self.parameter_widget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/MeshBoundaryLayer.ui"
        )
        QtCore.QObject.connect(
            self.parameter_widget.isb_number_of_layers,
            QtCore.SIGNAL("valueChanged(int)"),
            self.number_of_layers_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_min_thickness,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.min_thickness_changed,
        )
        # be careful of signal signature for QDoubleSpinbox
        QtCore.QObject.connect(
            self.parameter_widget.dsb_growth_rate,
            QtCore.SIGNAL("valueChanged(double)"),
            self.growth_rate_changed,
        )
        self.init_parameter_widget()

        # geometry selection widget
        # start with Solid in list!
        self.selection_widget = selection_widgets.GeometryElementsSelection(
            obj.References, ["Solid", "Face", "Edge", "Vertex"], True, False
        )

        # form made from param and selection widget
        self.form = [self.parameter_widget, self.selection_widget]

    def accept(self):
        self.set_mesh_boundarylayer_props()
        self.obj.References = self.selection_widget.references
        self.selection_widget.finish_selection()
        return super().accept()

    def reject(self):
        self.selection_widget.finish_selection()
        return super().reject()

    def init_parameter_widget(self):
        self.min_thickness = self.obj.MinimumThickness
        self.number_of_layers = self.obj.NumberOfLayers
        self.growth_rate = self.obj.GrowthRate
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_min_thickness).bind(
            self.obj, "MinimumThickness"
        )
        self.parameter_widget.qsb_min_thickness.setProperty("value", self.min_thickness)

        FreeCADGui.ExpressionBinding(self.parameter_widget.dsb_growth_rate).bind(
            self.obj, "GrowthRate"
        )
        self.parameter_widget.dsb_growth_rate.setProperty("value", self.growth_rate)

        FreeCADGui.ExpressionBinding(self.parameter_widget.isb_number_of_layers).bind(
            self.obj, "NumberOfLayers"
        )
        self.parameter_widget.isb_number_of_layers.setProperty("value", self.number_of_layers)

    def set_mesh_boundarylayer_props(self):
        self.obj.MinimumThickness = self.min_thickness
        self.obj.NumberOfLayers = self.number_of_layers
        self.obj.GrowthRate = self.growth_rate

    def min_thickness_changed(self, base_quantity_value):
        self.min_thickness = base_quantity_value

    def number_of_layers_changed(self, value):
        self.number_of_layers = value

    def growth_rate_changed(self, value):
        self.growth_rate = value



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM histogram plot task panel"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package task_post_histogram
#  \ingroup FEM
#  \brief task panel for post histogram plot

from PySide import QtCore, QtGui

import FreeCAD
import FreeCADGui

from . import base_fempostpanel
from femguiutils import extract_link_view as elv

translate = FreeCAD.Qt.translate


class _TaskPanel(base_fempostpanel._BasePostTaskPanel):
    """
    The TaskPanel for editing properties of glyph filter
    """

    def __init__(self, vobj):
        super().__init__(vobj.Object)

        # data widget
        self.data_widget = QtGui.QWidget()
        self.data_widget.show_table = QtGui.QPushButton()
        self.data_widget.show_table.setText(translate("FEM", "Show Table"))

        vbox = QtGui.QVBoxLayout()
        vbox.addWidget(self.data_widget.show_table)
        vbox.addSpacing(10)

        extracts = elv.ExtractLinkView(self.obj, False, self)
        vbox.addWidget(extracts)

        self.data_widget.setLayout(vbox)
        self.data_widget.setWindowTitle(translate("FEM", "Table Data"))
        self.data_widget.setWindowIcon(FreeCADGui.getIcon(":/icons/FEM_PostSpreadsheet.svg"))

        self.__init_widgets()

        # form made from param and selection widget
        self.form = [self.data_widget]

    # Setup functions
    # ###############

    def __init_widgets(self):

        # connect data widget
        self.data_widget.show_table.clicked.connect(self.showTable)

    @QtCore.Slot()
    def showTable(self):
        self.obj.ViewObject.Proxy.show_visualization()



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2026 FreeCAD Project Association
# SPDX-FileCopyrightText: 2025 Furgo
# SPDX-FileNotice: Part of the FreeCAD project.

################################################################################
#                                                                              #
#   FreeCAD is free software: you can redistribute it and/or modify            #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   FreeCAD is distributed in the hope that it will be useful,                 #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with FreeCAD. If not, see https://www.gnu.org/licenses       #
#                                                                              #
################################################################################

import FreeCAD as App
import Arch
import ArchSectionPlane
import Draft
from bimtests import TestArchBaseGui


class TestArchWindowGui(TestArchBaseGui.TestArchBaseGui):

    def test_change_window_opening(self):
        """Tests if changes to a window opening touches the window's chain of hosts"""

        # Create a wall, a window, a level and a section.
        points = [App.Vector(0.0, 0.0, 0.0), App.Vector(2000.0, 0.0, 0.0)]
        line = Draft.make_wire(points)
        wall = Arch.makeWall(line, height=2000)
        wpl = App.Placement(App.Vector(500, 0, 1500), App.Vector(1, 0, 0), -90)
        win = Arch.makeWindowPreset(
            "Open 1-pane",
            width=1000.0,
            height=1000.0,
            h1=50.0,
            h2=50.0,
            h3=50.0,
            w1=100.0,
            w2=50.0,
            o1=0.0,
            o2=50.0,
            placement=wpl,
        )
        win.Hosts = [wall]
        level = Arch.makeFloor()
        level.addObject(wall)
        section = Arch.makeSectionPlane(level)
        App.ActiveDocument.recompute()

        # Change opening from 0 to 50 (= 45 degrees):
        svg = ArchSectionPlane.getSVG(section)
        win.Opening = 50
        App.ActiveDocument.recompute()
        svg_new = ArchSectionPlane.getSVG(section)
        self.assertNotEqual(svg, svg_new)

        # Invert opening:
        svg = svg_new
        win.ViewObject.Proxy.invertOpening()
        App.ActiveDocument.recompute()
        svg_new = ArchSectionPlane.getSVG(section)
        self.assertNotEqual(svg, svg_new)

        # Invert hinge:
        svg = svg_new
        win.ViewObject.Proxy.invertHinge()
        App.ActiveDocument.recompute()
        svg_new = ArchSectionPlane.getSVG(section)
        self.assertNotEqual(svg, svg_new)



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD post extractors 2D"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package post_histogram
#  \ingroup FEM
#  \brief Post processing plot displaying lines

import FreeCAD

from . import base_fempostextractors
from . import base_fempythonobject

_PropHelper = base_fempythonobject._PropHelper

from vtkmodules.vtkCommonCore import vtkDoubleArray
from vtkmodules.vtkCommonDataModel import vtkTable
from vtkmodules.vtkCommonExecutionModel import vtkStreamingDemandDrivenPipeline

from PySide.QtCore import QT_TRANSLATE_NOOP


class PostFieldData2D(base_fempostextractors.Extractor2D):
    """
    A post processing extraction of two dimensional field data
    """

    ExtractionType = "Field"

    def __init__(self, obj):
        super().__init__(obj)

    def _get_properties(self):
        prop = [
            _PropHelper(
                type="App::PropertyBool",
                name="ExtractFrames",
                group="Multiframe",
                doc=QT_TRANSLATE_NOOP(
                    "FEM", "Specify if the field shall be extracted for every available frame"
                ),
                value=False,
            ),
        ]
        return super()._get_properties() + prop

    def execute(self, obj):

        # on execution we populate the vtk table
        table = vtkTable()

        if not obj.Source:
            obj.Table = table
            return

        dataset = obj.Source.getDataSet()
        if not dataset:
            obj.Table = table
            return

        timesteps = []
        if obj.ExtractFrames:
            # check if we have timesteps
            info = obj.Source.getOutputAlgorithm().GetOutputInformation(0)
            if info.Has(vtkStreamingDemandDrivenPipeline.TIME_STEPS()):
                timesteps = info.Get(vtkStreamingDemandDrivenPipeline.TIME_STEPS())
            else:
                FreeCAD.Console.PrintWarning(
                    'No frames available in data, ignoring "ExtractFrames" property'
                )

        if not timesteps:
            # get the dataset and extract the correct array
            xarray = self._x_array_from_dataset(obj, dataset)
            if xarray.GetNumberOfComponents() > 1:
                xarray.SetName(obj.XField + " (" + obj.XComponent + ")")
            else:
                xarray.SetName(obj.XField)

            self._x_array_component_to_table(obj, xarray, table)

            yarray = self._y_array_from_dataset(obj, dataset)
            if yarray.GetNumberOfComponents() > 1:
                yarray.SetName(obj.YField + " (" + obj.YComponent + ")")
            else:
                yarray.SetName(obj.YField)

            self._y_array_component_to_table(obj, yarray, table)

        else:
            algo = obj.Source.getOutputAlgorithm()
            for timestep in timesteps:
                algo.UpdateTimeStep(timestep)
                dataset = algo.GetOutputDataObject(0)

                xarray = self._x_array_from_dataset(obj, dataset)
                if xarray.GetNumberOfComponents() > 1:
                    xarray.SetName(f"X - {obj.XField} ({obj.XComponent}) - {timestep}")
                else:
                    xarray.SetName(f"X - {obj.XField} - {timestep}")
                self._x_array_component_to_table(obj, xarray, table)

                yarray = self._y_array_from_dataset(obj, dataset)
                if yarray.GetNumberOfComponents() > 1:
                    yarray.SetName(f"{obj.YField} ({obj.YComponent}) - {timestep}")
                else:
                    yarray.SetName(f"{obj.YField} - {timestep}")
                self._y_array_component_to_table(obj, yarray, table)

        # set the final table
        obj.Table = table


class PostIndexOverFrames2D(base_fempostextractors.Extractor2D):
    """
    A post processing extraction for two dimensional data with X always being the frames
    """

    ExtractionType = "Index"

    def __init__(self, obj):
        super().__init__(obj)

    def _get_properties(self):
        prop = [
            _PropHelper(
                type="App::PropertyInteger",
                name="Index",
                group="Data",
                doc=QT_TRANSLATE_NOOP(
                    "FEM", "Specify for which point index the data should be extracted"
                ),
                value=0,
            ),
        ]
        return super()._get_properties() + prop

    def _setup_x_component_property(self, obj, point_data):
        # override to only allow "Frames" as X data
        obj.XComponent = ["Not a vector"]

    def _setup_x_properties(self, obj, dataset):
        # override to only allow "Frames" as X data
        obj.XField = ["Frames"]

    def execute(self, obj):

        # on execution we populate the vtk table
        table = vtkTable()

        if not obj.Source:
            obj.Table = table
            return

        dataset = obj.Source.getDataSet()
        if not dataset:
            obj.Table = table
            return

        # check if we have timesteps (required!)
        timesteps = []
        info = obj.Source.getOutputAlgorithm().GetOutputInformation(0)
        if info.Has(vtkStreamingDemandDrivenPipeline.TIME_STEPS()):
            timesteps = info.Get(vtkStreamingDemandDrivenPipeline.TIME_STEPS())

        algo = obj.Source.getOutputAlgorithm()

        frame_x_array = vtkDoubleArray()
        frame_y_array = vtkDoubleArray()
        idx = obj.Index

        if timesteps:
            setup = False
            frame_x_array.SetNumberOfTuples(len(timesteps))
            frame_x_array.SetNumberOfComponents(1)
            for i, timestep in enumerate(timesteps):

                frame_x_array.SetTuple1(i, timestep)

                algo.UpdateTimeStep(timestep)
                dataset = algo.GetOutputDataObject(0)
                array = self._y_array_from_dataset(obj, dataset, copy=False)

                # safeguard for invalid access
                if idx < 0 or array.GetNumberOfTuples() - 1 < idx:
                    raise Exception(
                        f"Invalid index: {idx} is not in range 0 - {array.GetNumberOfTuples()-1}"
                    )

                if not setup:
                    frame_y_array.SetNumberOfComponents(array.GetNumberOfComponents())
                    frame_y_array.SetNumberOfTuples(len(timesteps))
                    setup = True

                frame_y_array.SetTuple(i, idx, array)

        else:
            frame_x_array.SetNumberOfTuples(1)
            frame_x_array.SetNumberOfComponents(1)
            frame_x_array.SetTuple1(0, 0)

            algo.Update()
            dataset = algo.GetOutputDataObject(0)
            array = self._y_array_from_dataset(obj, dataset, copy=False)

            # safeguard for invalid access
            if idx < 0 or array.GetNumberOfTuples() - 1 < idx:
                raise Exception(
                    f"Invalid index: {idx} is not in range 0 - {array.GetNumberOfTuples()-1}"
                )

            frame_y_array.SetNumberOfComponents(array.GetNumberOfComponents())
            frame_y_array.SetNumberOfTuples(1)
            frame_y_array.SetTuple(0, idx, array)

        frame_x_array.SetName("Frames")
        if frame_y_array.GetNumberOfComponents() > 1:
            frame_y_array.SetName(f"{obj.YField} ({obj.YComponent}) @Idx {obj.Index}")
        else:
            frame_y_array.SetName(f"{obj.YField} @Idx {obj.Index}")

        table.AddColumn(frame_x_array)
        self._y_array_component_to_table(obj, frame_y_array, table)

        # set the final table
        obj.Table = table



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM postprocessing visualization base ViewProvider"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package view_base_fempostvisualizations
#  \ingroup FEM
#  \brief view provider for post visualization object

import FreeCAD
import FreeCADGui


class VPPostVisualization:
    """
    A View Provider for visualization objects
    """

    def __init__(self, vobj):
        vobj.Proxy = self
        self._setup_properties(vobj)
        vobj.addExtension("Gui::ViewProviderGroupExtensionPython")

    def _setup_properties(self, vobj):
        pl = vobj.PropertiesList
        for prop in self._get_properties():
            if not prop.name in pl:
                prop.add_to_object(vobj)

    def _get_properties(self):
        return []

    def attach(self, vobj):
        self.Object = vobj.Object
        self.ViewObject = vobj

    def isShow(self):
        # Mark ourself as visible in the tree
        return True

    def getDisplayModes(self, obj):
        return ["Dialog"]

    def doubleClicked(self, vobj):

        guidoc = FreeCADGui.getDocument(vobj.Object.Document)

        # check if another VP is in edit mode and close it then
        if guidoc.getInEdit():
            FreeCADGui.Control.closeDialog()
            guidoc.resetEdit()

        # open task dialog
        guidoc.setEdit(vobj.Object.Name)

        # show visualization
        self.show_visualization()

        return True

    def unsetEdit(self, vobj, mode=0):
        FreeCADGui.Control.closeDialog()
        return True

    def updateData(self, obj, prop):
        # If the data changed we need to update the visualization
        if prop == "Table":
            self.update_visualization()

    def onChanged(self, vobj, prop):
        # for all property changes we need to update the visualization
        self.update_visualization()

    def childViewPropertyChanged(self, vobj, prop):
        # One of the extractors view properties has changed, we need to
        # update the visualization
        self.update_visualization()

    def dumps(self):
        return None

    def loads(self, state):
        return None

    # To be implemented by subclasses:
    # ################################

    def update_visualization(self):
        # The visualization data or any relevant view property has changed,
        # and the visualization itself needs to update to reflect that
        raise FreeCAD.Base.FreeCADError("Not implemented")

    def show_visualization(self):
        # Shows the visualization without going into edit mode
        raise FreeCAD.Base.FreeCADError("Not implemented")

    def get_next_default_color(self):
        # Returns the next default color a new object should use
        # Returns color in FreeCAD property notation (r,g,b,a)
        # If the relevant extractors do not have color properties, this
        # can stay unimplemented
        raise FreeCAD.Base.FreeCADError("Not implemented")



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD table view widget to visualize vtkTable"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package vtk_table_view
#  \ingroup FEM
#  \brief A Qt widget to show a vtkTable

from PySide import QtGui
from PySide import QtCore

import FreeCAD
import FreeCADGui

from vtkmodules.vtkIOCore import vtkDelimitedTextWriter

translate = FreeCAD.Qt.translate


class VtkTableModel(QtCore.QAbstractTableModel):
    # Simple table model. Only supports single component columns
    # One can supply a header_names dict to replace the table column names
    # in the header. It is a dict "column_idx (int)" to "new name"" or
    # "orig_name (str)" to "new name"

    def __init__(self, header_names=None):
        super().__init__()
        self._table = None
        if header_names:
            self._header = header_names
        else:
            self._header = {}

    def setTable(self, table, header_names=None):
        self.beginResetModel()
        self._table = table
        if header_names:
            self._header = header_names
        self.endResetModel()

    def rowCount(self, index):

        if not self._table:
            return 0

        return self._table.GetNumberOfRows()

    def columnCount(self, index):

        if not self._table:
            return 0

        return self._table.GetNumberOfColumns()

    def data(self, index, role):

        if not self._table:
            return None

        if role == QtCore.Qt.DisplayRole:
            col = self._table.GetColumn(index.column())
            return col.GetTuple(index.row())[0]

        return None

    def headerData(self, section, orientation, role):

        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            if section in self._header:
                return self._header[section]

            name = self._table.GetColumnName(section)
            if name in self._header:
                return self._header[name]

            return name

        if orientation == QtCore.Qt.Vertical and role == QtCore.Qt.DisplayRole:
            return section

        return None

    def getTable(self):
        return self._table


class VtkTableSummaryModel(QtCore.QAbstractTableModel):
    # Simple model showing a summary of the table.
    # Only supports single component columns

    def __init__(self):
        super().__init__()
        self._table = None

    def setTable(self, table):
        self.beginResetModel()
        self._table = table
        self.endResetModel()

    def rowCount(self, index):

        if not self._table:
            return 0

        return self._table.GetNumberOfColumns()

    def columnCount(self, index):
        return 2  # min, max

    def data(self, index, role):

        if not self._table:
            return None

        if role == QtCore.Qt.DisplayRole:
            col = self._table.GetColumn(index.row())
            range = col.GetRange()
            return range[index.column()]

        return None

    def headerData(self, section, orientation, role):

        if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
            return ["Min", "Max"][section]

        if orientation == QtCore.Qt.Vertical and role == QtCore.Qt.DisplayRole:
            return self._table.GetColumnName(section)

        return None

    def getTable(self):
        return self._table


class VtkTableView(QtGui.QWidget):

    def __init__(self, model):
        super().__init__()

        self.model = model

        layout = QtGui.QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        # start with the toolbar
        self.toolbar = QtGui.QToolBar()
        csv_action = QtGui.QAction(self)
        csv_action.triggered.connect(self.exportCsv)
        csv_action.setIcon(FreeCADGui.getIcon("Std_Export"))
        csv_action.setToolTip(translate("FEM", "Export to CSV"))
        self.toolbar.addAction(csv_action)

        copy_action = QtGui.QAction(self)
        copy_action.triggered.connect(self.copyToClipboard)
        copy_action.setIcon(FreeCADGui.getIcon("edit-copy"))
        shortcut = QtGui.QKeySequence(QtGui.QKeySequence.Copy)
        copy_action.setToolTip(
            translate("FEM", "Copy selection to clipboard ({})".format(shortcut.toString()))
        )
        copy_action.setShortcut(shortcut)
        self.toolbar.addAction(copy_action)

        layout.addWidget(self.toolbar)

        # now the table view
        self.table_view = QtGui.QTableView()
        self.table_view.setModel(model)
        self.model.modelReset.connect(self.modelReset)

        # fast initial resize and manual resizing still allowed!
        header = self.table_view.horizontalHeader()
        header.setResizeContentsPrecision(10)
        self.table_view.resizeColumnsToContents()

        layout.addWidget(self.table_view)
        self.setLayout(layout)

    @QtCore.Slot()
    def modelReset(self):
        # The model is reset, make sure the header visibility is working
        # This is needed in case new data was added
        self.table_view.resizeColumnsToContents()

    @QtCore.Slot(bool)
    def exportCsv(self, state):

        file_path, filter = QtGui.QFileDialog.getSaveFileName(
            None, translate("FEM", "Save as csv file"), "", "CSV (*.csv)"
        )
        if not file_path:
            FreeCAD.Console.PrintMessage(
                translate("FEM", "CSV file export aborted: no filename selected")
            )
            return

        writer = vtkDelimitedTextWriter()
        writer.SetFileName(file_path)
        writer.SetInputData(self.model.getTable())
        writer.Write()

    @QtCore.Slot()
    def copyToClipboard(self):

        sel_model = self.table_view.selectionModel()
        selection = sel_model.selectedIndexes()

        if len(selection) < 1:
            return

        copy_table = ""
        previous = selection.pop(0)
        for current in selection:

            data = self.model.data(previous, QtCore.Qt.DisplayRole)
            copy_table += str(data)

            if current.row() != previous.row():
                copy_table += "\n"
            else:
                copy_table += "\t"

            previous = current

        copy_table += str(self.model.data(selection[-1], QtCore.Qt.DisplayRole))
        copy_table += "\n"

        clipboard = QtGui.QApplication.instance().clipboard()
        clipboard.setText(copy_table)



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD

import Fem
import ObjectsFem

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Plate Buckling",
        "meshtype": "face",
        "meshelement": "Tria6",
        "constraints": ["displacement", "force"],
        "solvers": ["ccxtools"],
        "material": "solid",
        "equations": ["buckling"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.buckling_platebuckling import setup
setup()


See forum topic post:
https://forum.freecad.org/viewtopic.php?f=18&t=20217&start=110#p509935

"""
    )


def setup(doc=None, solvertype="ccxtools"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric object
    geom_obj = doc.addObject("Part::Plane", "Plate")
    geom_obj.Width = 6000
    geom_obj.Length = 8000
    doc.recompute()
    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # solver
    if solvertype == "ccxtools":
        solver_obj = ObjectsFem.makeSolverCalculiXCcxTools(doc, "CalculiXCcxTools")
        solver_obj.WorkingDir = ""
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
    if solvertype == "ccxtools":
        solver_obj.SplitInputWriter = False
        solver_obj.AnalysisType = "buckling"
        solver_obj.BucklingFactors = 10
        solver_obj.GeometricalNonlinearity = "linear"
        solver_obj.ThermoMechSteadyState = False
        solver_obj.MatrixSolverType = "default"
        solver_obj.IterationsControlParameterTimeUse = False
        solver_obj.BucklingFactors = 1
    analysis.addObject(solver_obj)

    # shell thickness
    thickness_obj = ObjectsFem.makeElementGeometry2D(doc, 50, "Thickness")
    analysis.addObject(thickness_obj)

    # material
    material_obj = ObjectsFem.makeMaterialSolid(doc, "Steel")
    mat = material_obj.Material
    mat["Name"] = "CalculiX-Steel"
    mat["YoungsModulus"] = "210000 MPa"
    mat["PoissonRatio"] = "0.30"
    material_obj.Material = mat
    analysis.addObject(material_obj)

    # constraints displacement
    con_disp_x = ObjectsFem.makeConstraintDisplacement(doc, "ConstraintDisplacement_X")
    con_disp_x.References = [(geom_obj, "Edge1")]
    con_disp_x.xDisplacement = 0
    con_disp_x.xFree = False
    analysis.addObject(con_disp_x)

    con_disp_y = ObjectsFem.makeConstraintDisplacement(doc, "ConstraintDisplacement_Y")
    con_disp_y.References = [(geom_obj, "Vertex1")]
    con_disp_y.yDisplacement = 0
    con_disp_y.yFree = False
    analysis.addObject(con_disp_y)

    con_disp_z = ObjectsFem.makeConstraintDisplacement(doc, "ConstraintDisplacement_Z")
    con_disp_z.References = [
        (geom_obj, "Edge1"),
        (geom_obj, "Edge2"),
        (geom_obj, "Edge3"),
        (geom_obj, "Edge4"),
    ]
    con_disp_z.zDisplacement = 0
    con_disp_z.zFree = False
    analysis.addObject(con_disp_z)

    # constraint force
    con_force = ObjectsFem.makeConstraintForce(doc, "ConstraintForce")
    con_force.References = [(geom_obj, "Edge3")]
    con_force.Force = "17162160 N"
    con_force.Reversed = True
    con_force.Direction = (geom_obj, ["Edge2"])
    analysis.addObject(con_force)

    # mesh
    from .meshes.mesh_buckling_plate_tria6 import create_nodes, create_elements

    fem_mesh = generate_mesh.mesh_from_existing(create_nodes, create_elements)
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.FemMesh = fem_mesh
    femmesh_obj.Shape = geom_obj
    femmesh_obj.SecondOrderLinear = False
    femmesh_obj.CharacteristicLengthMax = "300.0 mm"
    femmesh_obj.ElementDimension = "2D"

    doc.recompute()
    return doc
